@@ -14,7 +14,7 @@ except ImportError:
 from txpostgres import txpostgres
 
 from twisted.trial import unittest
-from twisted.internet import defer
+from twisted.internet import defer, reactor
 
 simple_table_schema = """
 CREATE TABLE simple (
@@ -71,6 +71,15 @@ class PollableThing(object):
         return 42
 
 
+class CancellablePollableThing(PollableThing):
+
+    cancelled = False
+
+    def cancel(self):
+        assert not self.cancelled
+        self.cancelled = True
+
+
 class FakeReactor(object):
     """
     A reactor that just counts how many things were added and removed.
@@ -104,7 +113,7 @@ class FakeWrapper(txpostgres._PollingMixin):
         return self._pollable
 
 
-class TxPostgresPollingMixingTestCase(Psycopg2TestCase):
+class TxPostgresPollingMixinTestCase(Psycopg2TestCase):
 
     def test_empty(self):
         """
@@ -197,6 +206,31 @@ class TxPostgresPollingMixingTestCase(Psycopg2TestCase):
         d = p.poll()
         return self.assertFailure(d, txpostgres.UnexpectedPollResult)
 
+    def test_cancel(self):
+        """
+        Cancelling a C{Deferred} returned from poll() proxies the cancellation
+        to the pollable and raises C{_CancelInProgress}.
+        """
+        p = FakeWrapper()
+        p._pollable = CancellablePollableThing()
+
+        d = p.poll()
+        self.assertRaises(txpostgres._CancelInProgress, d.cancel)
+        self.assertEquals(p._pollable.cancelled, True)
+
+    def test_noCancelSupport(self):
+        """
+        Cancelling a C{Deferred} returned from poll() with a pollable that does
+        not support cancelling just raises C{_CancelInProgress}.
+        """
+        p = FakeWrapper()
+        p._pollable = PollableThing()
+
+        d = p.poll()
+        self.assertRaises(txpostgres._CancelInProgress, d.cancel)
+        self.assertRaises(AttributeError, getattr, p._pollable, 'cancel')
+        self.assertRaises(AttributeError, getattr, p._pollable, 'cancelled')
+
 
 class TxPostgresConnectionTestCase(Psycopg2TestCase):
 
@@ -290,6 +324,18 @@ class TxPostgresConnectionTestCase(Psycopg2TestCase):
         d.addCallback(lambda _: conn.connect())
         return self.assertFailure(d, txpostgres.UnexpectedPollResult)
 
+    def test_openRunCloseOpen(self):
+        conn = txpostgres.Connection()
+        connargs = dict(user=DB_USER, password=DB_PASS,
+                        host=DB_HOST, database=DB_NAME)
+        d = conn.connect(**connargs)
+        d.addCallback(lambda _: conn.runQuery("select 1"))
+        # make sure the txpostgres.Cursor created by runQuery got closed,
+        # otherwise it will still be polled and will result in an error
+        d.addCallback(lambda _: conn.close())
+        d.addCallback(lambda _: conn.connect(**connargs))
+        return d.addCallback(lambda _: conn.close())
+
 
 class _SimpleDBSetupMixin(object):
 
@@ -494,7 +540,7 @@ class TxPostgresQueryTestCase(_SimpleDBSetupMixin, Psycopg2TestCase):
         d = self.conn.runInteraction(interaction)
         d.addCallback(lambda _: self.fail("No exception"))
 
-        def check_error(f):
+        def checkError(f):
             f.trap(txpostgres.RollbackFailed)
             original = f.value.originalFailure
             # original should reference the error that started all the mess
@@ -509,7 +555,7 @@ class TxPostgresQueryTestCase(_SimpleDBSetupMixin, Psycopg2TestCase):
             self.assertEquals(errors[0].value.args[0], "boom")
             # restore or we won't be able to clean up the mess
             mp.restore()
-        d.addErrback(check_error)
+        d.addErrback(checkError)
 
         # rollback for real, or tearDown won't be able to drop the table
         return d.addCallback(lambda _: self.conn.runOperation("rollback"))
@@ -644,3 +690,59 @@ class TxPostgresConnectionPoolHotswappingTestCase(Psycopg2TestCase):
         def simple(c):
             self.assertRaises(ValueError, pool.remove, c._connection)
         return d.addCallback(lambda pool: pool.runInteraction(simple))
+
+
+class TxPostgresCancellationTestCase(_SimpleDBSetupMixin, Psycopg2TestCase):
+
+    def setUp(self):
+        # check for cancellation support in psycopg2, skip if not present
+        if not getattr(psycopg2._psycopg.connection, "cancel", None):
+            raise unittest.SkipTest(
+                "psycopg2 does not have query cancellation support. "
+                "You need at least version 2.3.0 of psycopg2 "
+                "to use query cancellation.")
+
+        return _SimpleDBSetupMixin.setUp(self)
+
+    def test_simpleCancellation(self):
+        d = self.conn.runQuery("select pg_sleep(1000)")
+        reactor.callLater(0, self.conn.cancel, d)
+        return self.failUnlessFailure(d, defer.CancelledError)
+
+    def test_directCancellation(self):
+        d = self.conn.runQuery("select pg_sleep(1000)")
+
+        def tryDirectCancel(d):
+            self.assertRaises(txpostgres._CancelInProgress, d.cancel)
+        reactor.callLater(0, tryDirectCancel, d)
+
+        return self.failUnlessFailure(d, defer.CancelledError)
+
+    def test_cancelInteraction(self):
+        def interaction(c):
+            d = c.execute("insert into simple values (1)")
+            d.addCallback(lambda c: c.execute("insert into simple values (2)"))
+            d.addCallback(lambda c: c.execute("select pg_sleep(1000)"))
+            return d.addCallback(lambda _: "interaction done")
+
+        d = self.conn.runInteraction(interaction)
+        # Wow, that 0.5 sure is ugly. But what can happen is that if we cancel
+        # too early, we can zap the wrong query (i.e. one of the inserts) and
+        # the pg_sleep will continue... See runInteraction's docstring.
+        reactor.callLater(0.5, self.conn.cancel, d)
+
+        d = self.failUnlessFailure(d, defer.CancelledError)
+        d.addCallback(lambda _: self.conn.runQuery(
+                "select * from simple"))
+        return d.addCallback(self.assertEquals, [])
+
+    def test_cancelMultipleQueries(self):
+        d1 = self.conn.runQuery("select pg_sleep(1000)")
+        d2 = self.conn.runQuery("select pg_sleep(1000)")
+        reactor.callLater(0, self.conn.cancel, d1)
+        reactor.callLater(0, self.conn.cancel, d2)
+
+        d1 = self.failUnlessFailure(d1, defer.CancelledError)
+        d2 = self.failUnlessFailure(d2, defer.CancelledError)
+
+        return defer.gatherResults([d1, d2])
