@@ -12,7 +12,7 @@ from psycopg2 import extensions
 from zope.interface import implements
 
 from twisted.internet import interfaces, reactor, defer
-from twisted.python import log
+from twisted.python import failure, log
 
 
 try:
@@ -30,6 +30,11 @@ class UnexpectedPollResult(Exception):
     Polling returned an unexpected result.
     """
 
+class _CancelInProgress(Exception):
+    """
+    A query cancellation is in progress.
+    """
+
 
 class _PollingMixin(object):
     """
@@ -71,7 +76,9 @@ class _PollingMixin(object):
             the pollable reaches the OK state.
         """
         if not self._pollingD:
-            self._pollingD = defer.Deferred()
+            self._pollingD = defer.Deferred(self._cancel)
+            # transform a psycopg2 QueryCanceledError into CancelledError
+            self._pollingD.addErrback(self._handleCancellation)
         ret = self._pollingD
 
         try:
@@ -116,6 +123,21 @@ class _PollingMixin(object):
             d, self._pollingD = self._pollingD, None
             d.errback(reason)
 
+    def _cancel(self, d):
+        try:
+            self.pollable().cancel()
+        except AttributeError:
+            # the pollable has no cancellation support, ignore
+            pass
+        # prevent Twisted from errbacking the deferred being cancelled, because
+        # the PostgreSQL protocol requires finishing the entire polling process
+        # before reusing the connection
+        raise _CancelInProgress()
+
+    def _handleCancellation(self, f):
+        f.trap(psycopg2.extensions.QueryCanceledError)
+        return failure.Failure(defer.CancelledError())
+
     # forward all other access to the underlying connection
     def __getattr__(self, name):
         return getattr(self.pollable(), name)
@@ -307,7 +329,8 @@ class Connection(_PollingMixin):
     def _runQuery(self, *args, **kwargs):
         c = self.cursor()
         d = c.execute(*args, **kwargs)
-        return d.addCallback(lambda c: c.fetchall())
+        d.addCallback(lambda c: c.fetchall())
+        return d.addBoth(lambda ret: (c.close(), ret)[1])
 
     def runOperation(self, *args, **kwargs):
         """
@@ -328,14 +351,15 @@ class Connection(_PollingMixin):
     def _runOperation(self, *args, **kwargs):
         c = self.cursor()
         d = c.execute(*args, **kwargs)
-        return d.addCallback(lambda _: None)
+        d.addCallback(lambda _: None)
+        return d.addBoth(lambda ret: (c.close(), ret)[1])
 
     def runInteraction(self, interaction, *args, **kwargs):
         """
         Run commands in a transaction and return the result.
 
         The 'interaction' is a callable that will be passed a
-        C{pgadbapi.Cursor} object. Before calling 'interaction' a new
+        C{txpostgres.Cursor} object. Before calling 'interaction' a new
         transaction will be started, so the callable can assume to be running
         all its commands in a transaction. If 'interaction' returns a
         C{Deferred} processing will wait for it to fire before proceeding.
@@ -360,7 +384,7 @@ class Connection(_PollingMixin):
 
         @type interaction: Any callable
         @param interaction: A callable whose first argument is a
-            L{pgadbapi.Cursor}.
+            L{txpostgres.Cursor}.
 
         @rtype: C{Deferred}
         @return: A Deferred that will file with the return value of
@@ -381,22 +405,41 @@ class Connection(_PollingMixin):
             # maybeDeferred in case cursor.execute raises a synchronous
             # exception
             e = defer.maybeDeferred(cursor.execute, "rollback")
-            def just_panic(rf):
+            def justPanic(rf):
                 log.err(rf)
                 return defer.fail(RollbackFailed(self, f))
             # if rollback failed panic
-            e.addErrback(just_panic)
+            e.addErrback(justPanic)
             # reraise the original failure afterwards
             return e.addCallback(lambda _: f)
         d.addCallback(commitAndPassthrough, c)
         d.addErrback(rollbackAndPassthrough, c)
+        d.addBoth(lambda ret: (c.close(), ret)[1])
 
         return d
 
+    def cancel(self, d):
+        """
+        Cancel the current operation. The cancellation does not happen
+        immediately, because the PostgreSQL protocol requires that the
+        application waits for confirmation after the query has been cancelled.
+        Cancelling an interaction is tricky, because if the interaction
+        includes sending multiple queries to the database server, you can't
+        really be sure which one are you cancelling.
+
+        @type d: C{Deferred}
+        @param d: a Deferred returned by one of L{txpostgres.Connection}
+            methods.
+        """
+        try:
+            d.cancel()
+        except _CancelInProgress:
+            pass
+
 
 class ConnectionPool(object):
     """
-    A poor man's pool of L{pgadbapi.Connection} instances.
+    A poor man's pool of L{txpostgres.Connection} instances.
 
     @type min: C{int}
     @ivar min: The amount of connections that will be open at start. The pool
@@ -540,7 +583,7 @@ class ConnectionPool(object):
         Run commands in a transaction and return the result.
 
         The 'interaction' is a callable that will be passed a
-        C{pgadbapi.Cursor} object. Before calling 'interaction' a new
+        C{txpostgres.Cursor} object. Before calling 'interaction' a new
         transaction will be started, so the callable can assume to be running
         all its commands in a transaction. If 'interaction' returns a
         C{Deferred} processing will wait for it to fire before proceeding.
@@ -562,7 +605,7 @@ class ConnectionPool(object):
 
         @type interaction: Any callable
         @param interaction: A callable whose first argument is a
-            L{pgadbapi.Cursor}.
+            L{txpostgres.Cursor}.
 
         @rtype: C{Deferred}
         @return: A Deferred that will file with the return value of
